#include "classes.h"
#include "configreader.h"

Simulator::Simulator() {

}

//*** setup

void Simulator::loadTracks(string fname) {
	//file syntax:
	//
	// id length [end connectid [end connectid...]]
	//
	TrackSegment* tr;
	vector<int> v;
	vector<int> connections;
	int c;
	ostringstream tmps;
	ConfigReader cf(fname," ");

	v.clear();

	while(cf.getEntry(v) != -1) {
		if(v.size() >= 2) {
			tr = new TrackSegment(v[0], v[1]);
			tracks[tr->getId()] = tr;
			//store connections for later
			c = 1;
			while(v.size() >= c+2) {
				// 3 entries: track, end, connectrack
				connections.push_back(v[0]);
				connections.push_back(v[++c]);
				connections.push_back(v[++c]);
			}
			//report
			tmps << "loaded track " << tr->getId() << " length " << tr->getLength();
			logprint(tmps.str());
			tmps.str("");
		}
		v.clear();
	}

	vector<int>::iterator con = connections.begin();
	while(con != connections.end()) {
		tracks[*con]->connect(*(con+1), tracks[*(con+2)]);
		//report
		tmps << "connect track " << *con << "[" << *(con+1) << "] to track " << *(con+2);
		logprint(tmps.str());
		tmps.str("");
		con += 3;
	}

}

void Simulator::loadSensors(string fname) {
	//file syntax:
	//
	// id trackid pos
	//

	Sensor* sn;
	vector<int> v;
	ostringstream tmps;
	ConfigReader cf(fname," ");

	v.clear();

	while(cf.getEntry(v) != -1) {
		if(v.size() == 3) {
			sn = new Sensor(v[0]);
			tracks[v[1]]->add(sn, v[2]);
			sensors[sn->getId()] = sn;
			//report
			tmps << "loaded sensor " << sn->getId() << " at " << v[2] << " on track " << v[1];
			logprint(tmps.str());
			tmps.str("");
		}
		v.clear();
	}

}

void Simulator::loadTurnOuts(string fname) {
	//file syntax:
	//
	// id rootid rootend posid posend negid negend
	//
	// neg should be connected for off(), pos for on()

	ostringstream tmps;

	TurnOut* tn;
	vector<int> v;
	ConfigReader cf(fname," ");

	v.clear();

	while(cf.getEntry(v) != -1) {
		if(v.size() == 7) {
			tn = new TurnOut(v[0], tracks[v[1]], v[2], tracks[v[3]], v[4], tracks[v[5]], v[6]);
			turnouts[tn->getId()] = tn;
			//report
			tmps << "loaded turnout "<<tn->getId();
			tmps << " root  "<<v[1]<<"["<<v[2]<<"]";
			tmps << " pos  "<<v[3]<<"["<<v[4]<<"]";
			tmps << " neg  "<<v[5]<<"["<<v[6]<<"]";
			logprint(tmps.str());
			tmps.str("");
		}
		v.clear();
	}
}

void Simulator::loadTrains(string fname) {
	//file syntax:
	//
	// carid trainid length trackid heading closerdist [speedmult]
	//
	Train* tr;
	TrainCar* car;
	TrackRange* pos;

	ostringstream tmps;
	vector<int> v;
	ConfigReader cf(fname," ");

	v.clear();

	while(cf.getEntry(v) != -1) {
		if(v.size() >= 6) {
			//find train if exists.
			map<int,Train*>::iterator findtr = trains.find(v[1]);
			if(findtr == trains.end()) {
				//create it!
				tr = new Train(v[1]);
				trains[tr->getId()] = tr;
			} else
				//got it!
				tr = (findtr->second);

			car = new TrainCar(v[0], v[2], v[4], tr);
			pos = new TrackRange(v[5], v[5]+car->getLength(), car->getLength(), tracks[v[3]]->getLength());
			tracks[v[3]]->add(car, pos, v[4]);
			tr->addCar(car);
		}
		if(v.size() == 7)
			tr->setSpeedMult(v[6]);
		else
			//default to 0.5
			tr->setSpeedMult(0.5);
		tmps << "loaded traincar " << car->getId() << " len " << v[2] << " on " << v[3] << " at " << v[5];
		logprint(tmps.str());
		tmps.str("");
		v.clear();
	}
}

//*** DCC/CCR commands

int Simulator::motor(int trainid, int dir, int speed) {
	ostringstream tmps;
	tmps << "train " << trainid << " got speed " << speed << " going " << dir;
	logprint(tmps.str());

	map<int,Train*>::iterator tn;
	tn = trains.find((int)trainid);
	if(tn == trains.end()) return 0; //not found.
	Train* train = (tn->second);
	train->setDirection(dir);
	train->setSpeed(speed);
	return 1;
}

int Simulator::emerStop(int trainid, int dir) {
	ostringstream tmps;
	tmps << "train " << trainid << " got emerstop ";
	logprint(tmps.str());

	map<int,Train*>::iterator tn;
	tn = trains.find((int)trainid);
	if(tn == trains.end()) return 0; //not found.
	Train* train = (tn->second);
	train->setDirection(dir);
	train->setSpeed(0);
	return 1;
}

int Simulator::kill() {
	map<int,Train*>::iterator p = trains.begin();
	while(p != this->trains.end()) {
		(p->second)->setSpeed(0);
		p++;
	}
	return 1;
}

int Simulator::reset() {
	return 1;
}

short Simulator::TO_On(short id) {
	map<int,TurnOut*>::iterator tn;
	tn = turnouts.find((int)id);
	if(tn == turnouts.end()) return 0; //not found.

	(tn->second)->on();
	return 1;
}

short Simulator::TO_Off(short id) {
	map<int,TurnOut*>::iterator tn;
	tn = turnouts.find((int)id);
	if(tn == turnouts.end()) return 0; //not found.

	(tn->second)->off();
	return 1;
}

short Simulator::TO_Toggle(short id) {
	map<int,TurnOut*>::iterator tn;
	tn = turnouts.find((int)id);
	if(tn == turnouts.end()) return 0; //not found.

	(tn->second)->toggle();
	return 1;
}
short Simulator::Get_TO_Status(short id) {
	map<int,TurnOut*>::iterator tn;
	tn = turnouts.find((int)id);
	if(tn == turnouts.end()) return 0; //not found.

	return (short) (tn->second)->getStatus();
}

short Simulator::Get_Sensor_Status(short id) {
	map<int,Sensor*>::iterator sn;
	sn = sensors.find((int)id);
	if(sn == sensors.end()) return 0; //not found.

	return (short) (sn->second)->getStatus();
}

//*** run one movement of trains
void Simulator::step() {
	map<int,Train*>::iterator p = trains.begin();
	while(p != this->trains.end()) {
		(p->second)->move();
		p++;
	}
}

//*** graphical access

int Simulator::getCarPositions(int trackid, map<int,TrackRange*> &ret) {
	map<int,TrackSegment*>::iterator tn;
	tn = tracks.find((int)trackid);
	if(tn == tracks.end()) return 0; //not found.

	ret = (tn->second)->getCarPositions();
	return 1;
}


//*** debug

void Simulator::printStatus() {
	char tmp[80];
	string x;
	map<int,Train*>::iterator t = trains.begin();
	x = itoa(trains.size(),tmp,10);
	debugprint("I have "+x+" trains:");
	while(t != this->trains.end()) {
		x = itoa(t->first,tmp,10);
		x = x +" is "+itoa((t->second)->getId(),tmp,10);
		debugprint(x);
		t++;
	}
	map<int,TrackSegment*>::iterator p = tracks.begin();
	while(p != this->tracks.end()) {
		(p->second)->printStatus();
		p++;
	}
	bool stat;
	int i;
	map<int,Sensor*>::iterator s = sensors.begin();
	while(s != this->sensors.end()) {
		stat = (s->second)->getStatus();
		i = (s->second)->getId();
		x = itoa(i,tmp,10);
		debugprint("Photocell "+x, (int)stat);
		s++;
	}
	map<int,TurnOut*>::iterator g = turnouts.begin();
	while(g != this->turnouts.end()) {
		stat = (g->second)->getStatus();
		i = (g->second)->getId();
		x = itoa(i,tmp,10);
		debugprint("Turnout "+x, (int)stat);
		g++;
	}
}