#include "classes.h"

/* Constructors */

TrackSegment::TrackSegment(){
	this->length = -1;
	this->createHeadings();
}

TrackSegment::TrackSegment(int id, int length){
	this->id = id;
	this->length = length;
	this->createHeadings();
}


/* Internal */

void TrackSegment::createHeadings(){
	this->endHeadings.push_back(-1);
	this->endHeadings.push_back(1);
	this->endHeadings.push_back(-1);
	this->endHeadings.push_back(1);
}

void TrackSegment::checkInteractions(TrainCar* car, TrackRange* old, TrackRange* cur){
	//check Sensors
	int id,position;
	TrackRange* tpos;
	char tmp[80];

	map<int,Sensor*>::iterator p = this->sensors.begin();
	while(p != this->sensors.end()) {
		id = (p->second)->getId();
		position = this->pcPositions[id];
		if(old->includes(position) && !cur->includes(position))
			(p->second)->leaving(car);
		if(cur->includes(position))
			(p->second)->entering(car);
		p++;
	}
	//check cars
	map<int,TrainCar*>::iterator c = this->cars.begin();
	while(c != this->cars.end()) {
		id = (c->second)->getId();
		tpos = this->carPositions[id];
		//if overlapping and not the same car and not part of the same train
		if(cur->includes(tpos) && id != car->getId() && (c->second)->getParent() != car->getParent()) {
			sprintf(tmp, "found collision between traincars %d and %d", car->getId(), (c->second)->getId());
			logprint(tmp);
			(c->second)->derail();
			car->derail();
		}
		c++;
	}
}

void TrackSegment::removeCar(TrainCar* car){
	this->cars.erase(this->cars.find(car->getId()));
	this->carPositions.erase(this->carPositions.find(car->getId()));
	this->carDest.erase(this->carDest.find(car->getId()));
}

int TrackSegment::getOpposingEnd(int end){
	if(end == 0) {
		return 1;
	} else if(end == 1) {
		return 0;
	} else if(end == 2) {
		return 3;
	} else {
		return 2;
	}
}

/* Interface */

void TrackSegment::connect(int end, TrackSegment* track){
	this->connections[end] = track;
	this->rconnections[track] = end;
}

void TrackSegment::disconnect(TrackSegment* track){
	map<int,TrackSegment*>::iterator c = this->connections.begin();
	while(c != this->connections.end()) {
		if((c->second) == track)
			this->connections.erase(c);
			this->rconnections.erase(track);
			break;
	}
}

void TrackSegment::add(TrainCar* car, TrackRange* pos, int dest){
	ostringstream tmps;
	tmps << "track " << id << " got " << car->getId() << " heading to " << dest;
	logprint(tmps.str());

	this->cars[car->getId()] = car;
	this->carPositions[car->getId()] = pos;
	this->carDest[car->getId()] = dest;
	car->addTrack(this);
}

void TrackSegment::add(Sensor* pc, int location){
	this->sensors[pc->getId()] = pc;
	this->pcPositions[pc->getId()] = location;
}

void TrackSegment::move(TrainCar* car, int dist){

	TrackRange* oldpos;
	TrackRange* pos = this->carPositions[car->getId()];
	oldpos = new TrackRange(pos);

	ostringstream tmps;
	tmps << "traincar " << car->getId() << " moving " << dist << " on track " << id;
	logprint(tmps.str());

	//this will give a negative value if heading towards a zero-end.
	int real_dist = dist * this->endHeadings[this->carDest[car->getId()]];

	//move the trackrange for the car.
	int over = pos->move(real_dist);

	if(over != 0) {
		map<int,TrackSegment*>::iterator newtrack;
		newtrack = connections.find(carDest[car->getId()]);
		if(newtrack == connections.end()) {
			//no connecting track in that direction.
			car->derail();
		} else {
			//pass to connecting track
			this->connections[carDest[car->getId()]]->accept(car, this, over);
		}
	}
	if(pos->isEmpty()){
		this->removeCar(car);
		car->removeTrack(this);

		//sprintf(tmp, "traincar %d left track %d", car->getId(), id);
		//logprint(tmp);
	} else {
		//sprintf(tmp, "traincar %d finished at %s on track %d", car->getId(), pos->printable().c_str(), id);
		//logprint(tmp);
	}

	this->checkInteractions(car, oldpos, pos);
}

void TrackSegment::switchHeading(TrainCar* car){
	int old = this->carDest[car->getId()];
	switch(old){
	case 0:
		this->carDest[car->getId()] = 1;
		break;
	case 1:
		this->carDest[car->getId()] = 0;
		break;
	}
}

void TrackSegment::accept(TrainCar* car, TrackSegment* from, int dist){
	//if we already have the car, ignore
	if((cars.find(car->getId())->second) == car) return;

	int fromEnd = this->rconnections[from];
	TrackRange* pos;
	int len = this->getLength();

	if(this->endHeadings[fromEnd] > 0)
		pos = new TrackRange(len-1, len, car->getLength(), len);
	else 
		pos = new TrackRange(0,1, car->getLength(),  len);
	
	this->add(car, pos, this->getOpposingEnd(fromEnd));

	if(dist > 1){
		this->move(car, dist-1);
	}
}

int TrackSegment::getLength(){
	return this->length;
}

int TrackSegment::getId() {
	return id;
}

map<int,TrackRange*> TrackSegment::getCarPositions() {
	return carPositions;
}

void TrackSegment::printStatus() {
	char tmp[80];
	int id;
	string x;
	debugprint("Track",this->id);
	map<int,TrainCar*>::iterator c = this->cars.begin();
	while(c != this->cars.end()) {
		id = (c->second)->getId();
		x = itoa(id,tmp,10);
		x = "Car " + x;
		x = x + " at " + carPositions[id]->printable();
		debugprint(x);
		c++;
	}
}