#ifndef CLASSES_H
#define CLASSES_H

//disable silly debug stl warnings
#pragma warning(disable : 4786)

#define DEBUG 0

#include <vector>
#include <map>
#include <string>
#include <sstream>
#include <iostream>

using namespace std;

class Simulator;
class TrackRange;
class TrackSegment;
class TurnOut;
class Train;
class TrainCar;
class Sensor;


//*** output globals
void debugprint(string txt);
void debugprint(string txt, int num);

void logprint(string,bool newline=1);
void logprint(int,bool newline=1);
void logprint(double,bool newline=1);
//***


/** Old forward-definitions

// for TrainCar
vector<TrackSegment*>;

// for Train
vector<TrainCar*>;

// for TrackSegment
map<int,TrainCar*>;
map<int,TrackRange*>;
map<int,int>
map<int,Sensor*>;
vector<int>;
map<int,TrackSegment*>;
map<TrackSegment*,int>;
*/


//
// Simulator
//

class Simulator
{
public:
	Simulator();

	//*** setup
	void loadTracks(string fname);
	void loadSensors(string fname);
	void loadTurnOuts(string fname);
	void loadTrains(string fname);
	//*** DCC/CCR commands
	int motor(int trainid, int dir, int speed);
	int emerStop(int trainid, int dir);
	int kill();
	int reset();
	short TO_On(short id);
	short TO_Off(short id);
	short TO_Toggle(short id);
	short Get_TO_Status(short id);
	short Get_Sensor_Status(short id);
	
	//*** run one movement of trains
	void step();
	//*** graphical access
	int getCarPositions(int trackid, map<int,TrackRange*> &ret);
	//*** debug
	void printStatus();

private:
	map<int,TrackSegment*> tracks;
	map<int,Sensor*> sensors;
	map<int,TurnOut*> turnouts;
	map<int,Train*> trains;
};

/*********************
  March 13, 2002 Luke Opperman

  This is a limited range class for TrackSegments
  Specialized for use as a "shadow" of a TrainCar
  on a track, this Range is typically a fixed size
  unless it hits 0 or max.

*********************/

//
// TrackRange
//

class TrackRange
{
public:
	TrackRange();
	TrackRange(TrackRange* t);
	TrackRange(int start, int end, int len, int max);
	
	//this could be operator overload for + etc,
	//but there's really only one case right now.
	int move(int dist);

	bool includes(int pos);
	bool includes(TrackRange* tr);
	bool isEmpty(); 

	int getStart();
	int getEnd();

	string printable();

protected:
	int start;
	int end;
	int len;
	int max;
};

//
// TrackSegment
//

class TrackSegment
{
public:
	TrackSegment();
	TrackSegment(int id, int length);
	
	void connect(int end, TrackSegment* track);
	void disconnect(TrackSegment* track);

	void add(TrainCar* car, TrackRange* pos, int dest);
	void add(Sensor* pc, int location);

	void move(TrainCar* car, int dist);
	void switchHeading(TrainCar* car);
	void accept(TrainCar* car, TrackSegment* from, int dist);
	void printStatus();
	map<int,TrackRange*> getCarPositions();
	int getLength();
	int getId();

protected:
	void createHeadings();
	void checkInteractions(TrainCar* car, TrackRange* old, TrackRange* cur);
	void removeCar(TrainCar* car);
	int getOpposingEnd(int end);

	map<int,TrainCar*> cars; //car id -> car object
	map<int,TrackRange*> carPositions; //car id -> TrackRange
	map<int,int> carDest; //car id -> end

	map<int,Sensor*> sensors; //pc id -> pc
	map<int,int> pcPositions; // pc id -> position
	
	vector<int> endHeadings; //end -> heading multiplier (1,-1)
	map<int,TrackSegment*> connections; //end -> connecting track
	map<TrackSegment*,int> rconnections; //connecting track -> end
	
	int length;
	int id;
};


//
// TurnOut
//
//
/* turnouts don't care about the actual curvature,
 just make sure pos is the track for on(), neg for off()

 tn.off()  
    \      /  pos
     \    / 
  neg \  /  tn.on()
       \/
        |
        |
        | root
*/

class TurnOut
{
public:
	TurnOut();
	TurnOut(int, TrackSegment*, int, TrackSegment*, int, TrackSegment*, int);

	void toggle();
	void on();
	void off();
	bool getStatus();
	int getId();

private:
	TrackSegment* root;
	int rootend;
	TrackSegment* pos;
	int posend;
	TrackSegment* neg;
	int negend;
	bool status;
	int id;
};


/*********************
  March 13, 2002 Luke Opperman

  Train consists of one or more TrainCars,
  it receives move commands from Simulator

*********************/

//
// Train
//

class Train
{
public:
	Train();
	Train(int id);
	
	int getId();
	void move();
	void derail();

	void addCar(TrainCar* car);
	void removeCar(TrainCar* car);
	void removeCar(int carId);

	void setSpeed(int speed);
	void setDirection(int dir);
	void setSpeedMult(double mult);

	int getLength();

protected:
	vector<TrainCar*> cars;
	int id;
	int speed;
	double speed_mult;
	int dir;
	bool derailed;
};

/*********************
  March 13, 2002 Luke Opperman

  TrainCars occupy space on at least one track,
  can be more than one track. Part of larger Train.

*********************/

//
// TrainCar
//

class TrainCar
{
public:
	TrainCar();
	TrainCar(int id, int len, int dir, Train* parent);
	
	int getId();
	void move();
	void derail();

	void setSpeed(int speed);
	void setDirection(int dir);
	void setParent(Train* p);
	void setSpeedMult(double mult);

	Train* getParent();
	int getLength();
	bool isDerailed();

	void addTrack(TrackSegment* track);
	void removeTrack(TrackSegment* track);

protected:
	vector<TrackSegment*> tracks;
	Train* parent;
	int id;
	int length;
	int speed;
	int dir;
	double speed_mult;
	double distfraction;
	bool derailed;
};

/*********************
  March 13, 2002 Luke Opperman

  Sensors occupy a single unit on a Track
  They interact with TrainCars (via Track)

*********************/

//
// Sensor
//

class Sensor
{
public:
	Sensor();
	Sensor(int id);
	
	int getId();
	void entering(TrainCar* car);
	void leaving(TrainCar* car);

	bool getStatus();

protected:
	int id;
	// status: true = covered.
	bool status;
};


#endif
